# -*- coding: utf-8 -*-

"""
Module for computing cloud-mask from a sen2cor clouds classification raster.
"""

import logging
from pathlib import Path
import pyproj
from shapely.geometry import box, Polygon, MultiPolygon, GeometryCollection, mapping, shape
from shapely.ops import transform
import rasterio
from rasterio.features import shapes, rasterize
import numpy as np

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def max_dtype_value(dtype):
    """Returns the numpy dtype maximum value."""
    return np.iinfo(dtype).max

def mask_footprint(raster_band, footprint, out_path="./footprint_masked.tif"):
    """Mask ."""
    with rasterio.open(str(raster_band), "r") as src:
        profile = src.profile
        band = src.read(1)

        fp_poly = Polygon(footprint)
        projection = lambda x, y: pyproj.transform(pyproj.Proj(init='epsg:4326'),
                                                   pyproj.Proj(profile["crs"].to_string()), x, y
                                                  )
        fp_poly_reproj = transform(projection, fp_poly)

        fp_band = rasterize(shapes=[mapping(fp_poly_reproj)],
                            all_touched=True,
                            dtype=np.uint8,
                            out_shape=(profile["width"], profile["height"]),
                            transform=profile["affine"]
                           )

        nodata_value = max_dtype_value(profile["dtype"])
        band_masked = np.where(fp_band == 1, band, nodata_value).astype(np.uint16)

        profile.update(driver="Gtiff",
                       nodata=nodata_value,
                       tiled=False,
                       transform=profile["affine"])

    with rasterio.open(str(out_path), 'w', **profile) as out:
        out.write(band_masked, 1)


def katana(geometry, threshold, count=0):
    """
    Split a Polygon into two parts across it's shortest dimension
    recursively
    """
    # Credits: Joshua Arnott
    # https://snorfalorpagus.net/blog/2016/03/13/splitting-large-polygons-for-faster-intersections/
    LOGGER.debug("Splitting polygons.")

    bounds = geometry.bounds
    width = bounds[2] - bounds[0]
    height = bounds[3] - bounds[1]

    if max(width, height) <= threshold or count == 250:
        # either the polygon is smaller than the threshold, or the maximum
        # number of recursions has been reached
        return [geometry]

    if height >= width:
        # split left to right
        part_1 = box(bounds[0], bounds[1], bounds[2], bounds[1]+height/2)
        part_2 = box(bounds[0], bounds[1]+height/2, bounds[2], bounds[3])

    else:
        # split top to bottom
        part_1 = box(bounds[0], bounds[1], bounds[0]+width/2, bounds[3])
        part_2 = box(bounds[0]+width/2, bounds[1], bounds[2], bounds[3])

    result = []
    for each in (part_1, part_2,):
        intersec = geometry.intersection(each)
        if not isinstance(intersec, GeometryCollection):
            intersec = [intersec]
        for part in intersec:
            if isinstance(part, (Polygon, MultiPolygon)):
                result.extend(katana(part, threshold, count+1))

    if count > 0:
        return result

    # convert multipart into singlepart
    final_result = []
    for geom in result:
        if isinstance(geom, MultiPolygon):
            final_result.extend(geom)
        else:
            final_result.append(geom)
    return final_result



def erosion_dilatation(features, threshold=2500, erosion=-20, dilatation=100):
    """
    erosion - dilatation

    :param :
    :param :
    :param :
    :param :
    """
    for feat in features:
        geom = shape(feat["geometry"])

        for part in katana(geom.buffer(0.0), threshold, count=0):
            # buffer(0.0) avoids ring self-intersection error
            eroded = shape(part).buffer(float(erosion))
            new_geom = eroded if int(eroded.area) else "POLYGON EMPTY"

            if str(new_geom) != "POLYGON EMPTY":  # small geometries are empty after erosion
                yield ({"properties": feat["properties"],
                        "geometry": mapping(new_geom.buffer(float(dilatation)))
                       })


def create_cloud_mask(cloud_mask, out_path="./cloud_mask.tif", buffering=True, \
    erosion=None, dilatation=None):
    """
    create cloud mask
    uint8

    :param :
    :param :
    :param :
    :param :
    :param :
    """
    LOGGER.info("Creating cloud-mask.")
    # normalisation
    with rasterio.open(str(cloud_mask), "r") as src:
        profile = src.profile
        band = src.read(1)

    # normalisation
    #print("normalisation")
    mean = band.mean()
    std_dev = np.std(band)
    band_norm = (band - mean) / std_dev

    # positive array
    #print("positive array")
    band_norm01 = np.where(band_norm > 0, 1, 0)
    band_norm01 = band_norm01.astype(np.uint8, copy=False)

    # cloud-less array
    #print("cloud-less array")
    band_norm01_masked = np.ma.masked_where(band_norm01 == 1, band_norm01).mask

    with rasterio.open(str(out_path), 'w',
                       driver="Gtiff",
                       height=profile["height"],
                       width=profile["width"],
                       count=1,
                       dtype=np.uint8,
                       transform=profile["affine"],
                       crs=profile["crs"],
                       tiled=False) as out:

        # with erosion-dilatation
        if buffering:
            LOGGER.info("Erosion-dilatation.")

            # vectorization
            vectors = ({'properties': {'raster_val': v}, 'geometry': s} \
                for i, (s, v) in enumerate(shapes(band_norm01,
                                                  mask=band_norm01_masked,
                                                  transform=profile["affine"])))

            # buffering
            buffers = erosion_dilatation(vectors,
                                         erosion=erosion,
                                         dilatation=dilatation)
            shapes_values = ((f["geometry"], 1) for f in buffers)

            # rasterisation
            burned = rasterize(shapes=shapes_values,
                               all_touched=False,
                               dtype=np.uint8,
                               out_shape=(profile["width"], profile["height"]),
                               transform=profile["affine"])

            # write band
            out.write(burned, 1)

        # without erosion dilatation
        else:
            out.write(band_norm01, 1)
            LOGGER.info("Computed cloud-mask at: {}.".format(str(out_path)))

    return str(Path(str(out_path)).absolute)

    # cascaded union
    #from shapely.ops import cascaded_union
    #print("cascaded union")
    #dissolved = cascaded_union( [shape(f["geometry"]) for f in buffers] )
    #shapes = ((mapping(g), 1) for i, g in enumerate(dissolved))  # with cascaded union

    ###### TEST ######
    #with fiona.open(str("vectors.shp"), "w",
                    #driver="ESRI Shapefile",
                    #crs=profile["crs"],
                    #schema={"geometry":"Polygon", "properties": {"raster_val": "int"}}) as dst:
        #dst.writerecords(vectors)
    #sys.exit(0)
    ##################
