# -*- coding: utf-8 -*-

"""
Module for computing colormaps from single band rasters.
"""

import logging
import csv
from pathlib import Path
from typing import Dict, Tuple
# import colorcet as cc
import rasterio
from rasterio.warp import calculate_default_transform, reproject, Resampling
from matplotlib.colors import LinearSegmentedColormap
import numpy as np


LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


# TODO: add colormap.lut to matplotlib colormap

#def colorcet_colormap_to_rgb(cmap, revers=False):
    #"""
    #"""
    #if revers:
        #d = reversed(cmap)
    #else:
        #d = cmap
    #d2 = {i: hex_to_rgb(hex_string) for i, hex_string in enumerate(d)}

    #return d2

#def hex_to_rgb(hex_string):
    #"""
    #Converts a hexidecimal code to a RGB value.

    #:param hex_string: hexidecimal string.
    #"""
    #hexs = hex_string.lstrip('#')
    #size = len(hexs)
    #r, g, b = tuple(int(hexs[i:i + size // 3], 16) for i in range(0, size, size // 3))
    #return r, g, b

def matplotlib_colormap_to_rgb(cmap, revers=False):
    """
    # https://matplotlib.org/examples/color/colormaps_reference.html
    """
    def float_to_integer(flt):
        """ """
        return int(round(flt*255))

    cmap_list = [cmap(i) for i in range(0,256)]
    if revers:
        cmap_list = reversed(d)

    cmap_dict = {i: (float_to_integer(v[0]), float_to_integer(v[1]), float_to_integer(v[2]))
                for i, v in enumerate(cmap_list)}
    return cmap_dict


def get_tsv_colormap(lut_path: str) -> Dict[int, Tuple[int, int, int]]:
    """
    reads a tsv look-up table (values separated by tabs) and returns
    the content as a dictionnary : {i:(r, g, b), ...}.

    :param lut_path: look-up table path.
    """
    LOGGER.debug("retrieving look-up table from {}".format(lut_path))

    if not Path(lut_path).exists():
        raise IOError("File not found!")

    with open(str(lut_path), "r") as csv_file:
        reader = csv.reader(csv_file, delimiter=" ")
        colormap_dict = dict()
        for row in reader:
            color = tuple([int(row[1]), int(row[2]), int(row[3])])
            colormap_dict[int(row[0])] = color
    return colormap_dict


def create_colormap(raster: str, lut_dict: Dict[int, Tuple[int, int, int]],
                    cloud_mask: str=None, out_path: str="./colormap.tif",
                    out_crs: str=None, clouds_color: str="black",
                    out_resolution: Tuple[int, int]=(100, 100)) -> str:
    """
    Colormapping of a single-band raster with a look-up table passed as
    a dictionnary {i:(r, g, b), ...}. If a cloud mask is provided, clouds
    are colored in black

    :param raster: description.
    :param lut_dict: description.
    :param cloud_mask: description.
    :param out_path: description.
    :param out_crs: description.
    :param clouds_color: description.
    :param out_resolution: description.
    """
    LOGGER.info("creating colormap")

    # cloud mask
    if not Path(str(cloud_mask)).is_file():
        cloud_mask = None
        LOGGER.error("Cloud mask path wrong or not provided \
        \nClouds will not be colored")

    # clouds color
    if clouds_color.lower() == "black":
        lut_dict[0] = (0, 0, 0)
    elif clouds_color.lower() == "white":
        lut_dict[0] = (255, 255, 255)
    else:
        LOGGER.warning('Available clouds colors: "black" or "white" \
        \nApplying default: black')

    # check out_crs
    if out_crs:
        try:
            out_crs = CRS.from_string(out_crs.upper())
        except Exception as e:
            out_crs = None
            print("Invalid CRS: {}\nUsing source raster CRS".format(e))


    with rasterio.open(str(raster)) as src:
        raster_band = src.read(1)
        profile = src.profile

        if cloud_mask:
            with rasterio.open(str(cloud_mask)) as cld_src:
                clouds_band = cld_src.read(1)

                # resample cloud_mask to raster grid
                cld_reproj = np.empty(raster_band.shape, dtype=np.uint8)
                reproject(source=clouds_band,
                          destination=cld_reproj,
                          src_transform=cld_src.affine,
                          src_crs=cld_src.crs,
                          dst_transform=src.affine,
                          dst_crs=src.crs,
                          resampling=Resampling.nearest)

                # clouds
                raster_band = np.where(cld_reproj == 0, raster_band, 32767)

        #band_mask_borders = np.where(raster_band != 32767, raster_band, -10000)

        cmap = np.where(raster_band == 32767, -10000, raster_band)
        cmap = (128 * (cmap/10000 + 1) * ((cmap+10000) > 0)).astype(np.uint8)

        # compute default transform, width and height to fit the out resolution
        dst_transform, dst_width, dst_height = calculate_default_transform(
            src.crs,
            src.crs,
            src.width,
            src.height,
            *src.bounds,
            resolution=out_resolution)

        out_crs = src.crs if not out_crs else out_crs
        profile.update(nodata=None,
                       dtype=np.uint8,
                       transform=dst_transform,
                       width=dst_width,
                       height=dst_height,
                       crs=out_crs)

    # write colormap to out_path
    with rasterio.open(str(out_path), "w", **profile) as dst:
        dst.write(cmap, 1)
        dst.write_colormap(1, lut_dict)

    return str(Path(str(out_path)).absolute)
