# -*- coding: utf-8 -*-

"""
This module contains a class for building requests
"""

import logging
import pickle
import re
import sys
import fiona
from pathlib import Path
from datetime import datetime
from collections import OrderedDict
from sentinelsat import SentinelAPI
from fiona import collection
from osgeo import ogr
from shapely.geometry import shape, Point, Polygon, box, MultiPolygon
from shapely.wkt import loads
from shapely.ops import cascaded_union
from pprint import pprint
import itertools

from config import Config
#from geo_utils import get_tiles_from_point, get_tiles_from_bbox, get_tiles_from_file
from utils import grouper
from geo_utils import get_tiles_from_point, get_tiles_from_bbox, get_tiles_from_file

# présentation problème #
# il n'est pas possible avec l'API du scihub de faire une recherche
# par tuile des images antérieures à 20161205
# pour contourner ce problème, on utilisera la géométrie de la footprint
# de la tuile d'intérêt


functions_data_path = Path("../../share/data")
s2_tiles_index = functions_data_path / 'sentinel_2_index.geojson'
s2_tiles_index_dict = functions_data_path / "s2_tiles_index_dict.p"

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class DataRequest:
    """ classe pour récupérer les noms des produits
    """

    def __init__(self, start_date=None, end_date=None, land_only=False):

        logger.debug("DataRequest __init__")

        if start_date is None:
            self.start_date = datetime.strptime("20150629", "%Y%m%d")
        else:
            self.start_date = start_date

        if end_date is None:
            self.end_date = datetime.now()
        else:
            self.end_date = end_date

        self.land_only = land_only
        self.tiles_to_keep = None
        self.tiles_to_keep_geom = dict()
        self.products_list = {"aws":{}, "hubs":{}}


    #@staticmethod
    #def get_tiles_from_point(lon, lat):
        #"""
        #returns tiles containing a point (longitude, latitude)
        #"""
        #tiles_to_keep = set()
        #with fiona.open(str(s2_tiles_index), 'r') as tiles_index:
            #point_geom = Point(lon, lat)

            #for tile in tiles_index:
                #tile_geom = shape(tile['geometry'])
                #if tile_geom.contains(point_geom):
                    #tiles_to_keep.add(tile["properties"]["Name"])
        #return tiles_to_keep


    #@staticmethod
    #def get_tiles_from_bbox(lon_min, lat_min, lon_max, lat_max):
        #"""
        #returns S2 tiles intersecting a bbox
        #"""
        #tiles_to_keep = set()
        #with fiona.open(str(s2_tiles_index), 'r') as tiles_index:
            #bbox_geom = box(lon_min, lat_min, lon_max, lat_max)

            #for tile in tiles_index:
                #tile_geom = shape(tile['geometry'])
                #if tile_geom.intersects(bbox_geom):
                    #tiles_to_keep.add(tile["properties"]["Name"])

        #return tiles_to_keep


    #@staticmethod
    #def get_tiles_from_file(polygons_file):
        #"""
        #returns tiles intersecting geojson or shapefile features
        #"""
        #ds_t = ogr.Open(str(s2_tiles_index), 0)
        #lyr_t = ds_t.GetLayer(0)
        #ds_p = ogr.Open(str(polygons_file), 0)
        #lyr_p = ds_p.GetLayer(0)

        #tiles_to_keep = set()
        #for feat in lyr_p:
            #poly_geom = feat.GetGeometryRef()
            #lyr_t.SetSpatialFilter(poly_geom)

            #for filtered_t in lyr_t:
                #tiles_to_keep.add( filtered_t.GetField("Name") )
        #return tiles_to_keep


    @staticmethod
    def _serialise_s2_tiles_index():
        """
        save sentinel_2_index.geojson as a pickle file at first launch
        """
        tiles_dict = dict()

        with fiona.open(str(s2_tiles_index), "r") as tiles_index:
            for tile in tiles_index:
                name = tile["properties"]["Name"]
                geom = shape(tile['geometry'])
                tiles_dict[name] = geom

        with open(str(s2_tiles_index_dict), "wb") as pfile:
            pickle.dump(tiles_dict, pfile)


    def _get_tiles_geom(self):
        """
        return the geometries of a S2 tiles list as a multipolygon
        """

        try:
            assert s2_tiles_index_dict.exists()
        except AssertionError:
            DataRequest.serialise_s2_tiles_index()

        with open(str(s2_tiles_index_dict), "rb") as pfile:
            tiles_dict = pickle.load(pfile)

        for tile_kept in self.tiles_to_keep:
            try:
                self.tiles_to_keep_geom[tile_kept] = tiles_dict[tile_kept]
            except KeyError:
                print("Incorrect tile name: {}".format(tile_kept))
                sys.exit(1)


    def from_tiles(self, *tiles):
        """
        return
        """
        self.tiles_to_keep = [re.sub("^T", "", tile.upper()) for tile in set(tiles)]
        self._get_tiles_geom()
        self._make_request()
        return self.products_list


    def from_point(self, lon, lat):
        """
        tiles from a longitude / latitude point
        """
        self.tiles_to_keep = get_tiles_from_point(lon, lat, land_only=self.land_only)
        self._get_tiles_geom()
        self.make_request()
        return self.products_list


    def from_bbox(self, lon_min, lat_min, lon_max, lat_max):
        """
        tiles from a bbox
        """
        self.tiles_to_keep = get_tiles_from_bbox(lon_min, lat_min, lon_max, lat_max, land_only=self.land_only)
        self._get_tiles_geom()
        self._make_request()
        return self.products_list


    def from_file(self, vectors_file):
        """
        tiles from a vector file
        """
        geom_tiles = get_tiles_from_file(vectors_file, land_only=self.land_only)
        self.tiles_to_keep = list(set(itertools.chain.from_iterable(
                                                        geom_tiles.values()
                                                        )))
        self._get_tiles_geom()
        self._make_request()
        return self.products_list


    def _make_request(self):
        """
        lol
        """
        logger.debug("_make_request")

        tileddate = datetime.strptime("20161101", "%Y%m%d")
        start_date = self.start_date

        if not isinstance(self.start_date, datetime):
                start_date = datetime.strptime(start_date, "%Y%m%d")

        logger.info("start_date: {}".format(start_date))

        if start_date > tileddate:
            #sys.exit(0)
            self._make_request_tiled_only()
        else:
            #sys.exit(0)
            self._make_request_not_tiled()


    def _make_request_not_tiled(self):
        """
        request scihub api using sentinelsat
        """
        logger.debug("_make_request_not_tiled")

        relative_orbit = None
        cloud_cover_percentage = (0, 100)

        cloudcoverpercentage = cloud_cover_percentage if cloud_cover_percentage else (0,100)
        relativeorbitnumber = relative_orbit if relative_orbit else None

        if self.tiles_to_keep is None:
            raise ValueError("Query geometry not provided")
        else:
            print("Tile:", self.tiles_to_keep)

        api = SentinelAPI(Config().get("scihub_id"), Config().get("scihub_pwd"))
        products_from_hubs = dict()
        products_from_aws = dict()

        # remove water-only tiles when land-only parameter is enabled
        #if self.land_only:
            #with open(str(functions_data_path / "s2_tiles_land.p"), "rb") as pfile:
                #tiles_land = pickle.load(pfile)
                #water_tiles = []
                #for tile in self.tiles_to_keep.copy():
                    #if tile not in tiles_land:
                        #self.tiles_to_keep.remove(tile)
                        #water_tiles.append(tile)
                        #self.tiles_to_keep_geom.pop(tile)
            #print("Ignoring water-only tiles:", water_tiles)

        # query by group of 3 tiles,  otherwise getting error message
        # "Request URI too long" from scihub
        for tiles_to_keep_triplet, tiles_to_keep_triplet_geom in zip(grouper(self.tiles_to_keep, 3), grouper(self.tiles_to_keep_geom.values(), 3)):
            tiles_to_keep = [tile for tile in tiles_to_keep_triplet if tile]
            tiles_to_keep_geom = [geom for geom in tiles_to_keep_triplet_geom if geom]

            # build a multipolygon from tiles geom
            query_geom = MultiPolygon(tiles_to_keep_geom)
            logging.debug("query geometry:\n{}".format(query_geom))

            # scihub request
            products = OrderedDict()
            products = api.query(query_geom,
                                 date=(self.start_date, self.end_date),
                                 order_by="+endposition",
                                 platformname='Sentinel-2',
                                 #relativeorbitnumber=relativeorbitnumber,
                                 cloudcoverpercentage=cloud_cover_percentage)

            # save products list as a pandas dataframe
            products_df = api.to_dataframe(products)

            if products_df.empty:
                return
            # a products dictionnay for each server (AWS vs hubs)
            # fill each dictionnary depending on the acquisition date
            for index, row in products_df[["title", "beginposition", "footprint"]].iterrows():

                # start date of the tiled S2 collection on the scihub server
                tileddate = datetime.strptime("20161101", "%Y%m%d")
                img_title = row[0]
                img_date = row[1].to_pydatetime()
                img_footprint = loads(row[2])

                for tile_name, tile_geom in self.tiles_to_keep_geom.items():
                    # in case of duplicates on the server
                    if img_title not in self.products_list["hubs"].keys() and img_title not in self.products_list["aws"].keys():

                        # tiled products are downloaded on hubs
                        if re.match(r".*_T[0-9]{2}[A-Z]{3}_.*", img_title):
                            if tile_name in img_title:
                                self.products_list["hubs"][img_title] = {"date": img_date,
                                                                              "tile": tile_name}
                                continue
                            else:
                                continue

                         # non-tiled products will be downloaded on aws
                        else:
                            if tile_geom.intersects(img_footprint):
                                # if relative_orbit is given
                                #if relative_orbit:
                                    #relative_orbit_pattern = r".*_R{}_.*".format(relative_orbit)
                                    #if re.match(relative_orbit_pattern, img_title):
                                        #products_from_aws[img_title] = img_date
                                        #continue
                                    #else:
                                        #continue
                                self.products_list["aws"][img_title] = {"date": img_date,
                                                                             "tile": tile_name}

        #pprint(self.products_list)
        #sys.exit(0)

        # pprint dicts in chronological order
        print("\nFrom AWS")
        pprint(list(OrderedDict(sorted(self.products_list["aws"].items(),
                                key=lambda t: t[1]["date"]))))
        print("\nFrom hubs")
        pprint(list(OrderedDict(sorted(self.products_list["hubs"].items(),
                                key=lambda t: t[1]["date"]))))



### REQUÊTE POUR PRODUITS TUILÉS (moins lourde)

    def _make_request_tiled_only(self):
        """
        request scihub api using sentinelsat
        """
        logger.debug("_make_request_tiled_only")

        relative_orbit = None
        cloud_cover_percentage = None

        cloudcoverpercentage = cloud_cover_percentage if cloud_cover_percentage else (0,100)
        relativeorbitnumber = relative_orbit if relative_orbit else None

        if self.tiles_to_keep is None:
            raise ValueError("Query tiles not provided")
        else:
            print("Sentinel2 tiles:\n", self.tiles_to_keep)

        api = SentinelAPI(Config().get("scihub_id"), Config().get("scihub_pwd"))
        products_from_hubs = dict()
        products_from_aws = dict()

        # remove water-only tiles when land-only parameter is enabled
        if self.land_only:
            with open(str(s2_tiles_index_dict), "rb") as pfile:
                tiles_land = pickle.load(pfile)
                water_tiles = []
                for tile in self.tiles_to_keep.copy():
                    if tile not in tiles_land:
                        self.tiles_to_keep.remove(tile)
                        water_tiles.append(tile)
                        self.tiles_to_keep_geom.pop(tile)
            print("Ignoring water-only tiles:", water_tiles)

        query_kwargs = {
                'platformname': 'Sentinel-2',
                'cloudcoverpercentage': cloudcoverpercentage,
                'date': (self.start_date, self.end_date)}

        products = OrderedDict()
        for tile in self.tiles_to_keep:
            kw = query_kwargs.copy()
            #kw['tileid'] = tile  # products after 2017-03-31
            kw['filename'] = '*_T{}_*'.format(tile)
            pp = api.query(**kw)
            products.update(pp)

        # save products list as a pandas dataframe
        products_df = api.to_dataframe(products)

        if products_df.empty:
            return
            # a products dictionnay for each server (AWS vs hubs)
            # fill each dictionnary depending on the acquisition date
        for index, row in products_df[["title", "beginposition"]].iterrows():
            img_title = row[0]
            img_date = row[1].to_pydatetime()
            self.products_list["hubs"][img_title] = {"date": img_date,
                                                          "tile": re.findall("_T([0-9]{2}[A-Z]{3})_", img_title)[0]}

        # pprint dicts in chronological order
        print("\nFrom hubs")
        pprint(list(OrderedDict(sorted(self.products_list["hubs"].items(),
                                key=lambda t: t[1]["date"]))))




#instance_start_date = '20150101'
#instance_start_date = '20180701'
#instance_end_date = '20180716'

#instance_start_date, instance_end_date = '20160123', '20160424'
#x = 55.5331
#y = -21.1331

#bbox = (54.66796875, -21.647217065387817, 58.4033203125, -19.652934210612436)

#data = DataRequest(start_date=instance_start_date,
                   #end_date=instance_end_date,
                   #land_only=False).from_point(lon=x, lat=y)


#data = DataRequest(start_date=instance_start_date,
                   #end_date=instance_end_date,
                   #land_only=True).from_bbox(*bbox)


#data = DataRequest(start_date=instance_start_date,
                   #end_date=instance_end_date,
                   #land_only=True).from_file("/home/seas-oi/Documents/NDVI-MODIS-SENTINEL/SIG/sites_mada/34_sites_sentinelles.shp")
                   #land_only=True).from_file("/home/seas-oi/Téléchargements/map.geojson")



