# -*- coding: utf-8 -*-
"""
Module for manipulating Sentinel-2 tiles index.
"""


import logging
import pickle
# from pathlib import Path
from typing import List, Set, Dict
import fiona
from osgeo import ogr
from shapely.geometry import shape, Point, box

from .config import SHARED_DATA


LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# paths
TILES_INDEX = SHARED_DATA.get("tiles_index")
TILES_INDEX_DICT = SHARED_DATA.get("tiles_index_dict")


def serialise_tiles_index() -> None:
    """
    Save sentinel_2_index.geojson as a pickle file at first launch
    """
    tiles_dict = dict()

    with fiona.open(str(TILES_INDEX), "r") as tiles_index:
        for tile in tiles_index:
            name = tile["properties"]["Name"]
            geom = shape(tile['geometry'])
            tiles_dict[name] = geom

    with open(str(TILES_INDEX_DICT), "wb") as pfile:
        pickle.dump(tiles_dict, pfile)


#def get_tiles_geom(tiles=[]):
    #"""
    #return the geometries of a S2 tiles list as a multipolygon
    #"""
    #if not isinstance(tiles, list):
        #LOGGER.error("Tiles argument must be a list")
        #raise TypeError("Tiles argument must be a list")

    #try:
        #assert TILES_INDEX_DICT.exists()
    #except AssertionError:
        #loggin.debug("")
        #serialise_TILES_INDEX()

    #tiles_geom = dict()

    #with open(str(TILES_INDEX_DICT), "rb") as pfile:
        #tiles_dict = pickle.load(pfile)

    #for tile in tiles:
        #try:
            #tiles_geom[tile] = tiles_dict[tile]
        #except KeyError:
            #LOGGER.error("Incorrect tile name: {}".format(tile))
            #sys.exit(1)

    #return tiles_geom


def get_tiles_from_point(lon: float, lat: float, land_only: bool=False) -> Set[str]:
    """
    Returns tiles containing a point (longitude, latitude)

    :param lon: longitude.
    :param lat: latitude.
    :param land_only: keep only tiles that contain lands.
    """
    intersected_tiles = set()

    with fiona.open(str(TILES_INDEX), 'r') as tiles_index:
        point_geom = Point(lon, lat)
        if land_only:
            tiles_index = filter(lambda feat: feat["properties"]["land"] == "True", tiles_index)

        for tile in tiles_index:
            tile_geom = shape(tile['geometry'])
            if tile_geom.contains(point_geom):
                intersected_tiles.add(tile["properties"]["Name"])
    return intersected_tiles


def get_tiles_from_bbox(lon_min: float, lat_min: float, lon_max: float,
                        lat_max: float, land_only=False) -> Set[str]:
    """
    Returns S2 tiles intersecting a bbox.

    :param lon_min: longitude.
    :param lat_min: latitude.
    :param lon_max: longitude.
    :param lat_max: latitude.
    :param land_only: keep only tiles that contain lands.
    """
    intersected_tiles = set()

    with fiona.open(str(TILES_INDEX), 'r') as tiles_index:
        bbox_geom = box(lon_min, lat_min, lon_max, lat_max)
        if land_only:
            tiles_index = filter(lambda feat: feat["properties"]["land"] == "True", tiles_index)

        for tile in tiles_index:
            tile_geom = shape(tile['geometry'])
            if tile_geom.intersects(bbox_geom):
                intersected_tiles.add(tile["properties"]["Name"])

    return intersected_tiles


def get_tiles_from_file(vectors_file: str, land_only: bool=False) -> Dict[int, List[str]]:
    """
    Returns the intersected tiles for each feature in the vectors file.

    :param vectors_file: path to a file containing vectors geometries.
    :param land_only: return only tiles that contain lands.
    """
    # OGR is used for better performances
    drv_gpkg = ogr.GetDriverByName("GPKG")
    ds_t = drv_gpkg.Open(str(TILES_INDEX), 0)
    ds_gf = ogr.Open(str(vectors_file), 0)
    lyr_gf = ds_gf.GetLayer(0)
    lyr_t = ds_t.GetLayer(0)
    # filter on tiles with land
    if land_only:
        lyr_t.SetAttributeFilter("land = 'True'")

    #this dict will contain the geometry id as key and the list of
    #tiles it intersects as value
    intersected_tiles_dict = dict()
    i = 0
    for feat in lyr_gf:
        feat_geom = feat.GetGeometryRef()
        lyr_t.SetSpatialFilter(feat_geom)

        contained_in_tiles = []
        for filtered_t in lyr_t:
            tile = filtered_t.GetField("Name")
            contained_in_tiles.append(tile)

        intersected_tiles_dict[str(i)] = contained_in_tiles if contained_in_tiles else None
        i += 1

    return intersected_tiles_dict


