# -*- coding: utf-8 -*-

"""
Module for defining radiometric indices.
"""

import logging
from pathlib import Path
from abc import ABCMeta, abstractmethod, abstractproperty
from matplotlib import cm

from .indices_functions import (create_raw_ndvi, create_raw_ndwi,
                                create_masked_indice)
from .colormap import matplotlib_colormap_to_rgb, create_colormap

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class Indice(metaclass=ABCMeta):
    """Indice abstract base class. Every indice class must implement these
    class attributes and object methods.

    _name: Name of the indice. Must be unique and without spaces.
    _filename_template: Template of the indice filename output. Will be
    filled with the product's identifier and the file extention.
    eg: "{product_identifier}_INDICE_NAME{ext}"
    _ext: Extension of the indice filename. (eg: ".tif", ".jp2")

    process_indice(out_path, nodata_clouds, quicklook):
    """

    @abstractproperty
    def name(self):
        pass

    @abstractproperty
    def filename_template(self):
        pass

    @abstractproperty
    def ext(self):
        pass

    @abstractmethod
    def process_indice(self):
        pass


class Ndvi(Indice):
    """
    NDVI = (NIR-VIR) / (NIR+VIR)

    NIR: band 08
    VIR: band 04
    """

    name = "NDVI"
    filename_template = "{product_identifier}_NDVI{ext}"
    ext = ".tif"
    colormap = cm.RdYlGn
    # _colors_table_path = functions_data_path = Path("../share/data/RdYlGn.lut")

    def __init__(self, L2aProduct_obj):
        if L2aProduct_obj is None:
            raise ValueError("A L2aProduct object must be provided")
        else:
            self.l2a_product = L2aProduct_obj

        # output path
        self.out_path = None

        # filenames
        self.indice_stem = self.filename_template.format(product_identifier=self.l2a_product.identifier,
                                                         ext="")
        self.indice_filename = self.indice_stem + self.ext


    def process_indice(self, out_path, reprocess=False, nodata_clouds=False, quicklook=False):
        """ process  """
        self.out_path = out_path

        if (out_path / self.indice_filename).exists() and not reprocess:
            logger.info("{} already exists".format(self.indice_filename))
        else:
            create_raw_ndvi(nir_path=self.l2a_product.b08_10m,
                            vir_path=self.l2a_product.b04_10m,
                            out_path=(out_path / self.indice_filename))

        if nodata_clouds:

            if self.l2a_product.user_cloud_mask is None:
                raise ValueError("Cloud mask does not exist")
            masked_indice_filename = self.indice_stem + "_MASKED" + self.ext

            if (self.out_path / masked_indice_filename).exists() and not reprocess:
                logger.info("{} already exists".format(masked_indice_filename))
            else:
                create_masked_indice(raw_indice_path=(self.out_path / self.indice_filename),
                                     cloud_mask_path=self.l2a_product.user_cloud_mask,
                                     out_path=(self.out_path / masked_indice_filename))


        if quicklook:
            cmap = matplotlib_colormap_to_rgb(self.colormap, revers=False)

            quicklook_filename = self.indice_stem + "_QUICKLOOK.tif"
            if (self.out_path / quicklook_filename).exists() and not reprocess:
                logger.info("{} already exists".format(quicklook_filename))
            else:
                create_colormap(raster=(self.out_path / self.indice_filename),
                                cloud_mask=self.l2a_product.user_cloud_mask,
                                # lut_dict=get_colormap(self._colors_table_path.absolute()),
                                lut_dict=cmap,
                                out_path=(self.out_path / quicklook_filename))


class NdwiGao(Indice):
    """
    NDWI(Gao) = (NIR-SWIR) / (NIR+SWIR)

    NIR: band 08
    SWIR: band 11
    """

    name = "NDWI_GAO"
    filename_template = "{product_identifier}_NDWIg{ext}"
    ext = ".tif"
    colormap = cm.RdYlBu
    #colormap = get_tsv_colormap("../../share/data/RdYlGn.lut")


    def __init__(self, L2aProduct_obj):
        if L2aProduct_obj is None:
            raise ValueError("A L2aProduct object must be provided")
        else:
            self.l2a_product = L2aProduct_obj

        # output path
        self.out_path = None

        # filenames
        self.indice_stem = self.filename_template.format(product_identifier=self.l2a_product.identifier,
                                                     ext="")
        self.indice_filename = self.indice_stem + self.ext

    def process_indice(self, out_path, reprocess=False, nodata_clouds=False, quicklook=False):
        """ process  """
        self.out_path = out_path

        if (self.out_path / self.indice_filename).exists() and not reprocess:
            logger.info("{} already exists".format(self.indice_filename))
        else:
            create_raw_ndwi(nir_path=self.l2a_product.b08_10m,
                            swir_path=self.l2a_product.b11_20m,
                            out_path=(out_path / self.indice_filename))

        if nodata_clouds:
            if self.l2a_product.user_cloud_mask is None:
                raise ValueError("Cloud mask does not exist")
            masked_indice_filename = self.indice_stem + "_MASKED" + self.ext

            if (self.out_path / masked_indice_filename).exists() and not reprocess:
                logger.info("{} already exists".format(masked_indice_filename))
            else:
                create_masked_indice(raw_indice_path=(self.out_path / self.indice_filename),
                                     cloud_mask_path=self.l2a_product.user_cloud_mask,
                                     out_path=(self.out_path / masked_indice_filename))

        if quicklook:
            cmap = matplotlib_colormap_to_rgb(self.colormap, revers=False)

            quicklook_filename = self.indice_stem + "_QUICKLOOK.tif"
            if (self.out_path / quicklook_filename).exists() and not reprocess:
                logger.info("{} already exists".format(quicklook_filename))
            else:
                create_colormap(raster=(self.out_path / self.indice_filename),
                                cloud_mask=self.l2a_product.user_cloud_mask,
                                lut_dict=cmap,
                                out_path=(self.out_path / quicklook_filename))



class IndicesCollectionMeta(type):
    """Adds special methods to IndicesCollection class.

    """
    def __iter__(cls):
        return iter(cls._indices_classes)

    def __contains__(cls, indice):
        return indice in cls._indices_classes

    def __len__(cls):
        return len(cls._indices_classes)

    def __str__(cls):
        return str(cls.list)

    @property
    def available_indices(cls):
        return list(cls._indices_classes.keys())

    @property
    def list(cls):
        return list(cls._indices_classes.keys())

    def get_indice_cls(cls, indice):
        """
        Returns the indice class.

        :param indice: name of the indice
        """
        try:
            indice_cls = cls._indices_classes[indice]
        except KeyError:
            logger.error("Indice: {} is not defined".format(indice))
            return None
        else:
            return indice_cls

class IndicesCollection(metaclass=IndicesCollectionMeta):
    """Defined indices.

    """
    # TODO: Implement class as a singleton.

    _indices_classes = {cls.__dict__["name"]:cls for cls in Indice.__subclasses__()}
