# -*- coding: utf-8 -*-

"""
This module contains functions to compute radiometric indices.
"""

import logging
from pathlib import Path
import rasterio
from rasterio.warp import reproject, Resampling
import numpy as np

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def create_raw_ndvi(nir_path=None, vir_path=None, out_path="./raw_ndvi.tif"):
    """
    ndvi
    """
    LOGGER.info("creating raw NDVI (int16)")
    with rasterio.open(str(nir_path)) as nir_src, \
            rasterio.open(str(vir_path)) as vir_src:

        nir_profile = nir_src.profile
        # vir_profile = vir_src.profile

        np.seterr(divide='ignore', invalid='ignore')    # ignore warnings when dividing by zero
        nir = nir_src.read(1).astype(np.float32)
        vir = vir_src.read(1).astype(np.float32)
        ndvi = ((nir - vir) / (nir + vir)*10000).astype(np.int16)

        ndvi_masked = np.where(nir != 0, ndvi, 32767)

    nir_profile.update(driver="Gtiff",
                       compress="DEFLATE",
                       tiled=False,
                       dtype=np.int16,
                       nodata=32767,
                       transform=nir_src.affine)

    with rasterio.open(str(out_path), "w", **nir_profile) as dst:
        dst.write(ndvi_masked, 1)

    return Path(str(out_path)).absolute


def create_raw_ndwi(nir_path=None, swir_path=None, out_path="./raw_ndwi.tif"):
    """
    ndwi
    """
    LOGGER.info("creating raw NDWI (int16)")

    with rasterio.open(str(nir_path)) as nir_src, \
            rasterio.open(str(swir_path)) as swir_src:

        nir_profile = nir_src.profile
        # swir_profile = swir_src.profile

        nir = nir_src.read(1).astype(np.float32)
        swir = swir_src.read(1).astype(np.float32)

        # reproject swir band (20m) to nir band resolution (10m)
        swir_reproj = np.empty(nir.shape, dtype=np.float32)
        reproject(source=swir,
                  destination=swir_reproj,
                  src_transform=swir_src.affine,
                  src_crs=swir_src.crs,
                  dst_transform=nir_src.affine,
                  dst_crs=nir_src.crs,
                  resampling=Resampling.nearest)

        np.seterr(divide='ignore', invalid='ignore')    # ignore warnings when dividing by zero
        ndwi = ((nir - swir_reproj) / (nir + swir_reproj)*10000).astype(np.int16)

        ndwi_masked = np.where(nir != 0, ndwi, 32767)

    nir_profile.update(driver="Gtiff",
                       compress="DEFLATE",
                       tiled=False,
                       dtype=np.int16,
                       nodata=32767,
                       transform=nir_src.affine)

    with rasterio.open(str(out_path), "w", **nir_profile) as dst:
        dst.write(ndwi_masked, 1)

    return Path(str(out_path)).absolute



def create_masked_indice(raw_indice_path=None, cloud_mask_path=None, out_path="./masked_indice.tif"):
    """
    masked indice
    """
    # TODO: Improve the function in order to mask any raster.

    LOGGER.info("Cloud-masking indice (int16)")

    with rasterio.open(str(raw_indice_path)) as indice_src, \
            rasterio.open(str(cloud_mask_path)) as cld_src:

        profile = indice_src.profile
        raw_indice = indice_src.read(1)
        cld = cld_src.read(1)

        # repoject cloud_mask to ndvi resolution
        cld_reproj = np.empty(raw_indice.shape, dtype=np.uint8)
        reproject(source=cld,
                  destination=cld_reproj,
                  src_transform=cld_src.affine,
                  src_crs=cld_src.crs,
                  dst_transform=indice_src.affine,
                  dst_crs=indice_src.crs,
                  resampling=Resampling.nearest)

        indice_borders_mask = np.where(raw_indice > 0, raw_indice, 32767)
        indice_cloud_mask = np.where(cld_reproj == 0, indice_borders_mask, 32767)

        profile.update(nodata=32767,
                       transform=indice_src.affine)

    with rasterio.open(str(out_path), "w", **profile) as dst:
        dst.write(indice_cloud_mask, 1)

    return str(Path(str(out_path)).absolute)
