# -*- coding: utf-8 -*-

"""
Module for managing products and tiles in the library and temp folders.
"""

import os
import pathlib
import subprocess
import shutil
import functools
import logging
import re
import fiona

import xml.etree.ElementTree as et
from pprint import pprint
from pathlib import Path
from collections import OrderedDict, namedtuple
from datetime import datetime
from pprint import pformat
# type annotations
from typing import List, Set, Dict, Tuple, Optional

from .utils import str_to_datetime, grouper
from .config import Config, SHARED_DATA
from .xmlparser import MetadataParser
from .sen2cor import process_sen2cor
from .cloud_mask import create_cloud_mask
from .indices import IndicesCollection
from .products import L1cProduct, L2aProduct


s2_tiles_index = SHARED_DATA.get("tiles_index")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Library:
    """
    """
    _temp_path = Path(Config().get("temp_path"))
    _l1c_path = Path(Config().get("l1c_path"))
    _l2a_path = Path(Config().get("l2a_path"))
    _indices_path = Path(Config().get("indices_path"))

    def __init__(self) -> None:
        self._l1c = [f.name for f in self._l1c_path.glob("*")]
        self._l2a = [f.name for f in self._l2a_path.glob("*")]
        self._indices = dict()

        for indice in [f.name for f in self._indices_path.glob("*")]:
            self.__dict__[indice.lower()] = [f.name for f in self._indices_path.glob("{}/*".format(indice))]


    @property
    def l1c(self):
        return self._l1c

    @property
    def l2a(self):
        return self._l2a

    @property
    def indices(self):
        return self._indices


class LibraryContainer:
    """ Temporary container for downloading L1C and processing L2A.

    :param l1c_identifier:
    :param tile:
    """
    def __init__(self, identifier=None, tile: str = None) -> None:
        self._l1c_identifier = identifier.replace("L2A_","L1C_")
        self._l2a_identifier = identifier.replace("L1C_","L2A_")

        self.l1c = L1cProduct(self._l1c_identifier)
        self.l2a = L2aProduct(self._l2a_identifier)

        self.tile = self.l1c.tile


class TempContainer:
    """ Temporary folder for downloading L1C and processing L2A

    :param l1c_identifier:
    :param tile:
    """

    _temp_path = Path(Config().get("temp_path"))

    def __init__(self, l1c_identifier=None, tile: str=None) -> None:
        """
        décris moi
        """

        self.tile = tile
        self.temp_id = l1c_identifier + "_TEMP"
        self.temp_path = self._temp_path / self.temp_id

        self.l1c = L1cProduct(identifier=l1c_identifier, tile=tile, path=self.temp_path)

        #self.l2a_identifier = self.l1c.identifier.replace("L1C_","L2A_").replace("_USER_", "_OPER_")
        #self.l2a = None


    def create_temp_folder(self):
        """Creates a temporary folder."""
        logger.info("{}: creating temp folder".format(self.l1c.identifier))
        try:
            self.temp_path.mkdir(exist_ok=True)
        except FileNotFoundError:
            print("Invalid temp folder path:", self.temp_path)
            raise
        return self

    def unzip_l1c(self):
        """Unzip L1C zip file."""
        zip_file = self.temp_path / (self.l1c.identifier + ".zip")
        if zip_file.exists():
            logger.info("{} : Unzipping L1C archive".format(self.l1c.identifier))
            shutil.unpack_archive(str(zip_file), str(self.temp_path))
            os.remove(str(zip_file))
        return self

    #def process_l2a(self):
        #"""Process sen2cor."""
        #self.l1c.process_l2a()
        #self.l2a = L2aProduct(identifier=self.l2a_identifier,
                              #tile=self.tile,
                              #path=self.temp_path)
        #return self

    #def process_cloud_mask(self):
        #"""Process cloud mask."""
        #if self.l2a is None:
            #self.l2a = L2aProduct(identifier=self.l2a_identifier,
                                  #tile=self.tile,
                                  #path=self.temp_path)
        #self.l2a.process_cloud_mask()
        #return self

    #def process_indices(self,
            #indices_list: list, nodata_clouds: bool, quicklook: bool):
        #"""Process indices.

        #:param indices_list:
        #:param nodata_clouds:
        #:param quicklook:
        #"""
        #if self.l2a is None:
            #self.l2a = L2aProduct(identifier=self.l2a_identifier,
                                  #tile=self.tile,
                                  #path=self.temp_path)
        #self.l2a.process_indices(indices_list, nodata_clouds, quicklook)
        #return self

    #def archive_files(self):
        #"""Archives L1C and L2A files in the library."""
        #try:
            #self.l1c.archive()
        #except:
            #pass
        #try:
            #self.l2a.archive()
        #except:
            #pass
        #return self

    def delete_temp_folder(self):
        """Deletes temporary folder."""
        logger.info("Deleting temp folder: {}".format(self.temp_path))
        if not list(self.temp_path.glob("*.SAFE")):
            try:
                shutil.rmtree(str(self.temp_path))
            except FileNotFoundError:
                logger.warning("Temp folder does not exist: {}".format(self.temp_path))
        else:
            logger.error("Temp folder contains a SAFE product.")
        return self

