# -*- coding: utf-8 -*-

"""
Module for managing products and tiles in the library and temp folders.
"""

import pathlib
import logging
import re

from pprint import pprint
from pathlib import Path
from collections import OrderedDict, namedtuple
from datetime import datetime
from pprint import pformat
# type annotations
from typing import List, Set, Dict, Tuple, Optional

from .config import Config, SHARED_DATA
from .utils import str_to_datetime, grouper
from .indices import IndicesCollection
from .products import L1cProduct, L2aProduct


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


# namedtuple
TileProduct = namedtuple("TileProduct", ["identifier", "date", "cloud_cover"])

class ProductsList:
    """Class for managing tiles products lists.
    Used in Tile.
    """

    def __init__(self):
        self._dict = dict()
        self._time_index = dict()

    @property
    def products(self):
        """Returns products names."""
        return list(self._dict.keys())

    @property
    def dates(self):
        """Returns products dates."""
        return list(self._time_index.keys())

    @property
    def first(self):
        """Returns oldest product."""
        try:
            min_date = min(self.dates)
        except ValueError:
            return TileProduct(None, None, None)
        prod = self._time_index[min_date]
        return TileProduct(prod, min_date, self._dict[prod]["cloud_cover"])

    @property
    def last(self):
        """Returns newest product."""
        try:
            max_date = max(self.dates)
        except ValueError:
            return TileProduct(None, None, None)
        prod = self._time_index[max_date]
        return TileProduct(prod, max_date, self._dict[prod]["cloud_cover"])

    def filter_dates(self, date_min: datetime=None, date_max: datetime=None):
        """Filters products list in time range.

        :param date_min:
        :param date_max:
        """
        min_date = str_to_datetime(date_min, "ymd") if date_min else self.first.date
        max_date = str_to_datetime(date_max, "ymd") if date_max else self.last.date
        filtre = ProductsList()
        for k,v in self._dict.items():
            if v["date"] >= min_date and v["date"] <= max_date:
                filtre[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtre

    def filter_clouds(self, cover_min: int=0, cover_max: int=100):
        """Filters products list.

        :param cover_min:
        :param cover_max:
        """
        filtre = ProductsList()
        for k,v in self._dict.items():
            if v["cloud_cover"] >= cover_min and v["cloud_cover"] <= cover_max:
                filtre[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtre

    def __len__(self):
        return len(self._dict)

    def __contains__(self, item):
        return item in self._dict

    def __repr__(self):
        return pformat([self[k] for k in self.products], indent=1, width=1)

    def __iter__(self):
        return (self[k] for k in self._dict)

    def __getitem__(self, item):
        if item in self._dict:
            return TileProduct(item, self._dict[item]["date"], self._dict[item]["cloud_cover"])
        else:
            raise KeyError("{} not found".format(item))

    def __setitem__(self, item, value):
        self._dict[item] = {"date": value["date"], "cloud_cover": value["cloud_cover"]}
        self._time_index[value["date"]] = item


class IndicesList(ProductsList):
    """Used in Tile.
    """
    @property
    def raws(self):
        filtre = ProductsList()
        for k,v in self._dict.items():
            if not "MASK" in k and not "QUICKLOOK" in k:
                filtre[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtre

    @property
    def masks(self):
        filtre = ProductsList()
        for k,v in self._dict.items():
            if "MASK" in k:
                filtre[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtre

    @property
    def quicklooks(self):
        filtre = ProductsList()
        for k,v in self._dict.items():
            if "QUICKLOOK" in k:
                filtre[k] = {"date": v["date"], "cloud_cover": v["cloud_cover"]}
        return filtre


class Tile:
    """Manage tiles.

    :param name: tile's name.

    Usage:
        >>> Tile("40KCC")
    """
    def __init__(self, name: str) -> None:
        self.name = name
        self._paths = {"l1c": Path(Config().get("l1c_path")) / name,
                       "l2a": Path(Config().get("l2a_path")) / name,
                       "indices": {}
                       }
        self._indices_path = Path(Config().get("indices_path"))

        self._products = {"l1c": ProductsList(),
                          "l2a": ProductsList(),
                          "indices": dict()}

        self._get_indices_paths()
        self._get_l1c_list()
        self._get_l2a_list()
        self._get_indices_list()


    def _get_indices_paths(self) -> None:
        """Updates self._paths"""
        for indice in self._indices_path.glob("*/{}".format(self.name)):
            self._paths["indices"][str(indice.parent.name.lower())] = indice

    def _get_l1c_list(self) -> None:
        """
        """
        for f in self._paths["l1c"].glob("*L1C_*.SAFE"):
            cloud_cover = L1cProduct(identifier=f.name, tile=self.name, path=None).cloud_coverage_assessment
            self._products["l1c"][f.name] = {"date": Tile._get_date(f.name), "cloud_cover": float(cloud_cover)}

    def _get_l2a_list(self) -> None:
        """
        """
        for f in self._paths["l2a"].glob("*L2A*.SAFE"):
            l1c_name = f.name.replace("L2A_", "L1C_").replace("_USER_", "_OPER_")
            try:
                date = self._products["l1c"][l1c_name].date
                cloud_cover = self._products["l1c"][l1c_name].cloud_cover
            except KeyError:
                date = Tile._get_date(f.name)
                cloud_cover = L2aProduct(f.name, self.name).cloud_coverage_assessment
            self._products["l2a"][f.name] = {"date": date, "cloud_cover": float(cloud_cover)}

    def _get_indices_list(self):
        """
        """
        for indice, path in self._paths["indices"].items():
            if path.is_dir():
                self._products["indices"][indice] = IndicesList()
                indice_template = IndicesCollection.get_indice_cls(indice.upper()).filename_template
                indice_ext = IndicesCollection.get_indice_cls(indice.upper()).ext
                for f in path.glob("*{}".format(indice_ext)):
                    try:
                        indice_pattern = re.sub("{.*?}", "", indice_template)
                        remove_pattern = "{}.*".format(indice_pattern)
                        l2a_name = re.sub(remove_pattern, '', f.name) + ".SAFE"
                        date = self._products["l2a"][l2a_name].date
                        cloud_cover = self._products["l2a"][l2a_name].cloud_cover
                    except KeyError:
                        date = Tile._get_date(f.name)
                        cloud_cover = None
                    self._products["indices"][indice][f.name] = {"date": date, "cloud_cover": cloud_cover}
                self.__dict__[indice] = self._products["indices"][indice]

    @staticmethod
    def is_tile_name_valid(tile_name: str) -> bool:
        """Checks if the tile_name is a valid Sentinel-2 tile.

        :param tile_name: tile's name.
        """
        if not re.match(r"^[0-9]{2}[A-Z]{3}$", tile_name):
            return False
        with fiona.open(str(s2_tiles_index), 'r') as tiles_index:
            tiles_names = [tile["properties"]["Name"] for tile in tiles_index]
            return tile_name in tiles_names

    @staticmethod
    def _get_date(product_name: str) -> datetime:
        """
        """
        date_pattern = re.findall(r"[0-9]{8}T[0-9]{6}", product_name)
        if product_name.startswith("S2A_OPER"):
            date = date_pattern[1]
        else:
            date = date_pattern[0]
        return str_to_datetime(date, "filename")

    def __repr__(self):
        return "Tile('{}')".format(self.name)

    def __str__(self):
        return self.name

    @property
    def paths(self):
        return self._paths

    @property
    def l1c(self):
        return self._products["l1c"]

    @property
    def l2a(self):
        return self._products["l2a"]

    @property
    def l1c_missings(self):
        """
        """
        prods_list = ProductsList()
        missings_l1c_set = set(self.l2a.products) - {identifier.replace("L1C_", "L2A_").replace("__OPER__", "_USER_") \
                                                    for identifier in self.l1c.products}
        for prod in missings_l1c_set:
            prods_list[prod] = {"date": self._products["l2a"][prod].date,
                                "cloud_cover": self._products["l2a"][prod].cloud_cover}
        return prods_list

    @property
    def l2a_missings(self):
        """
        """
        prods_list = ProductsList()
        missings_l2a_set = set(self.l1c.products) - {identifier.replace("L2A_", "L1C_").replace("_USER_", "_OPER_") \
                          for identifier in self.l2a.products}
        for prod in missings_l2a_set:
            prods_list[prod] = {"date": self._products["l1c"][prod].date,
                                "cloud_cover": self._products["l1c"][prod].cloud_cover}
        return prods_list
